/**
 * Copyright (c) Members of the EGEE Collaboration. 2007-2010. 
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.  
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); 
 * you may not use this file except in compliance with the License. 
 * You may obtain a copy of the License at 
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0 
 * 
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. 
 * See the License for the specific language governing permissions and 
 * limitations under the License.
 *
 *
 *  Authors:
 *  2007-
 *     Oscar Koeroo <okoeroo@nikhef.nl>
 *     David Groep <davidg@nikhef.nl>
 *     NIKHEF Amsterdam, the Netherlands
 *     <grid-mw-security@nikhef.nl> 
 *
 *
 */


#include "ssl-common.h"
#include "scas_log.h"

/* network functions */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>



/* Is OpenSSL already initialized */
int openssl_is_initialized = 0;




/******************************************************************************
Function:     SSL_recv
Description:
              This function will use the SSL object to readout the SSL transport 
              receive buffer for num amount of bytes and put it to the data buffer.
              Error message handling included
              Note the entire SSL read buffer will be read, until it doesn't fit 
              in the buffer.
Parameters:
Returns:
     < 0  :  Error code from SSL_read()
    == 0  :  Nothing to be done
     > 0  :  Amount of bytes read from the buffer
******************************************************************************/
int SSL_recv (SSL * ssl, void * data, int num)
{
    int error_code = 0;
    int rc = 0;
    int nread = 0;
    int retry = 0;

    scas_log_debug (5, "%s: entering %s\n", __func__, __func__);

    if (!ssl)
    {
        scas_log (1, "%s: No SSL object presented to SSL_recv()\n", __func__);
        return -1;
    }

    /* clear buffer */
    bzero (data, num);

    /* rc is the return code when rc <= 0, otherwise its the amount of bytes read */

    /*
     * Igor's patch to be more resiliant to long distance connections 
     * which don't fill the buffers fast enough to let this work at once
     *
    rc = SSL_read (ssl, data, num);
    if (rc < 0)
        print_ssl_error_msg (ssl, rc, 1, &error_code); 
    */

    /* Endless retry (this might bite me) */
    while (1) 
    {
        rc = SSL_read (ssl, data, num);
        if (rc < 0) 
        {
            int unrecoverable = print_ssl_error_msg (ssl, rc, 1, &error_code);
            if (unrecoverable)
                break; /* fatal error */
            /* else retry */
        } 
        else 
        {
            break; /* no error */
        }
    } 


    scas_log_debug (1, "%s: Read %d bytes, buffer size is %d.\n", __func__, rc, num);
    return rc;


  /* Out of scope */
    while ((nread < num) && 
           (rc = SSL_read (ssl, (unsigned char *) data + nread, num - nread)) && 
           (retry < 5))
    {
        /* Got bytes */
        if (rc > 0)
        {
            nread += rc;
        }
        else
        {
            if (print_ssl_error_msg (ssl, rc, 1, &error_code))
                break;
            else 
            {
                retry++;
                scas_log_debug (3, "%s: continueing SSL_read(), retry %d\n", __func__, retry);
                usleep (50 * 1000 * retry);
                continue;
            } 
        }
    }
    scas_log_debug (1, "%s: Read %d bytes, buffer size is %d.\n", __func__, nread, num);

    print_ssl_error_msg (ssl, rc, 0, &error_code); 
    if ((error_code == SSL_ERROR_WANT_READ) || (error_code == SSL_ERROR_WANT_WRITE))
    {
        scas_log (3, "%s: Something is still in the transport buffer\n", __func__);
    }

    return nread ? nread : rc;
}



/******************************************************************************
Function:     SSL_server_accept
Description:
              This will take the open socket descriptor and the pre-setup CTX,
              connect the socket with the SSL layer and initiate the SSL_connect() 
              handshake 

Parameters:
Returns:
    >NULL : A pointer to ready to use SSL object pointer
    NULL  : FUBAR
******************************************************************************/
SSL * SSL_server_accept (int opensocket, char * remote_host, SSL_CTX * ctx)
{
    int    rc = 0;
    SSL *  ssl  = NULL;
    BIO *  bio  = NULL;
    int              i     = 0;
    int              depth = 0;
    int              ssl_error_code = 0;
    X509 *           ssl_peer_cert       = NULL;
    STACK_OF(X509) * ssl_peer_cert_chain = NULL;
    
    int activity = 0;
    fd_set readfds;
    struct timeval timeout;


    /* Creating the SSL object out of the CTX */
    if (ctx)
    {
        /* Creating the SSL layer */
        if (!(ssl = SSL_new(ctx)))
        {
            scas_log (1, "%s: Error: Failed to create the SSL layer\n", __func__);
            return NULL;
        }
        else
        {
            scas_log_debug (5, "%s: SSL object created from CTX\n", __func__);
            SSL_clear(ssl);
        }
    } 
    else
    {
        scas_log (1, "%s: Error: can't create SSL handle out of CTX structure\n", __func__);
        return NULL;
    }


    /* Connecting the Socket to the SSL layer */
    if (!(bio = BIO_new_socket (opensocket, BIO_NOCLOSE)))
    {
        scas_log (1, "%s: Error: Failed to tie the socket to a SSL BIO\n", __func__);
        SSL_free(ssl);
        return NULL;
    }
    else
    {
        scas_log_debug (5, "%s: BIO created from socket\n", __func__);
    }


    /* Tie the socket as a file descriptor to the SSL layer */
    if (! SSL_set_fd(ssl, opensocket))
    {
        scas_log (1, "%s: Error: SSL_set_fd() failed!\n", __func__);
        SSL_free(ssl);
        return NULL;
    }
    else
    {
        scas_log_debug (5, "%s: Passed setting binding fd to SSL\n", __func__);
    }


    /* Set the connection to non-blocking mode */
    /*
    BIO_set_nbio (bio, 1);
     */

    SSL_set_bio(ssl, bio, bio);
    scas_log_debug (5, "%s: BIO set for SSL\n", __func__);

    BIO_set_nbio(bio,1);
    scas_log_debug (5, "%s: BIO set to non-blocking\n", __func__);

/*
    SSL_set_accept_state(ssl);
    scas_log_debug (5, "%s: Passed SSL set accept state\n", __func__);
*/


/********************************************/
    timeout.tv_sec=0;
    timeout.tv_usec=500000;

    scas_log_debug (5, "%s: Waiting for client activity...\n", __func__);
    for (;;)
    {
        FD_ZERO(&readfds);

        /* setup which sockets to listen on */
        FD_SET(opensocket, &readfds);

        /* wait for connection, forever if we have to */
        activity = select(opensocket + 1, &readfds, (fd_set*)0, (fd_set*)0, &timeout);

        if ((activity < 0) && (errno!=EINTR))
        {
            /* there was an error with select() */
            SSL_free(ssl);
            return NULL;
        }
        else if (activity == 0)
        {
            /* Time out reached */
            scas_log_debug (3, "%s: Client SSL activity time out reached.\n", __func__);
            SSL_free(ssl);
            return NULL;
        }
        else
        {
            if (FD_ISSET(opensocket, &readfds))
            {
                scas_log_debug (5, "%s: Got client activity.\n", __func__);
                break;
            }
        }
    }

/********************************************/

    /* SSL accepting */
    scas_log_debug (5, "%s: Waiting in SSL_accept()\n", __func__);
    if ((rc = SSL_accept(ssl)) <= 0)
    {
        scas_log_debug (3, "%s: Warning: failure in SSL_accept()\n", __func__);
        if (print_ssl_error_msg (ssl, rc, 1, &ssl_error_code))
        {
            SSL_free(ssl);
            return NULL;
        } 
        else
        {
            if ((ssl_error_code == SSL_ERROR_WANT_READ) || (ssl_error_code == SSL_ERROR_WANT_WRITE))
            {
                scas_log (3, "%s: Something is still in the transport buffer\n", __func__);
                SSL_free(ssl);
                return NULL;
            }

            /* SSL accepting - retry */
            if ((rc = SSL_accept(ssl)) <= 0)
            {
                scas_log (1, "%s: Error: failure in SSL_accept() (already retried)\n", __func__);
                print_ssl_error_msg (ssl, rc, 1, NULL);
                SSL_free(ssl);
                return NULL;
            } 
        }
    }
    scas_log_debug (5, "Successful SSL_accept()\n");

/***/

    if (!(ssl_peer_cert = SSL_get_peer_certificate(ssl)))
    {
        scas_log (1, "%s: Error: Could not get peer certificate from the SSL handle.\n", __func__);
        SSL_free(ssl);
    }

    if (!(ssl_peer_cert_chain = SSL_get_peer_cert_chain(ssl)))
    {
        scas_log (1, "%s: Error: Could not get peer chain of certificates from the SSL handle\n", __func__);
        SSL_free(ssl);
    }
    else
    {
        /*
         * So we have a chain... but does it contain the peer certificate?
         * ... lets check and add it if not
         */

        depth = sk_X509_num(ssl_peer_cert_chain);
        scas_log_debug (5, "%s: Chain depth = %d\n", __func__, depth);
        sk_X509_insert(ssl_peer_cert_chain, ssl_peer_cert, 0);
        scas_log_debug (5, "%s: Added the peer certificate to the chain\n", __func__);
        depth = sk_X509_num(ssl_peer_cert_chain);
        scas_log_debug (5, "%s: Chain depth = %d\n", __func__, depth);

        for (i = 0; i < depth; i++)
        {
            char * oneliner_tmp = X509_NAME_oneline(X509_get_subject_name (sk_X509_value(ssl_peer_cert_chain, i)), NULL, 0);

            scas_log_debug (5, "%s: (%d) -> \"%s\" (%s) .\n", 
                               __func__, 
                               i, 
                               oneliner_tmp,
                               x509IsCA(sk_X509_value(ssl_peer_cert_chain, i)) ? "is CA" : "not a CA");
            free(oneliner_tmp);
        }
    }


    /* Connection stuff done, client connected */
    scas_log_debug (4, "%s: SSL Connection accepted and opened\n", __func__);

    return ssl;
}



/******************************************************************************
Function:     SSL_client_connect
Description:
              This will take the open socket descriptor and the pre-setup CTX,
              connect the socket with the SSL layer and initiate the SSL_connect() 
              handshake 

Parameters:
Returns:
    >NULL : A pointer to ready to use SSL object pointer
    NULL  : FUBAR
******************************************************************************/
SSL * SSL_client_connect (int opensocket, SSL_CTX * ctx) 
{
    int    rc = 0;
    SSL *  ssl  = NULL;
    BIO *  bio  = NULL;
    int    retry = 0;


    /* Creating the SSL object out of the CTX */
    if (ctx)
    {
        /* Creating the SSL layer */
        if (!(ssl = SSL_new(ctx)))
        {
            scas_log (1, "%s: Error: Failed to create the SSL layer\n", __func__);
            return NULL;
        }
        else
        {
            scas_log_debug (5, "%s: SSL object created from CTX\n", __func__);
        }
    }
    else
    {
        scas_log (1, "%s: Error: can't create SSL handle out of CTX structure\n", __func__);
        return NULL;
    }



    /* Connecting the Socket to the SSL layer */
    if (!(bio = BIO_new_socket (opensocket, BIO_NOCLOSE)))
    {
        scas_log (1, "%s: Error: Failed to tie the socket to a SSL BIO\n", __func__);
        SSL_free(ssl);
        return NULL;
    }
    else
    {
        scas_log_debug (5, "%s: BIO created from socket\n", __func__);
    }


    /* Set the connection to non-blocking mode */
    /* 
    BIO_set_nbio (bio, 1);
     */

    /* Binding/setting the BIO to the SSL layer */
    SSL_set_bio(ssl, bio, bio);
    scas_log_debug (5, "%s: BIO set for SSL\n", __func__);


    /* The retries are local to the system mostly, hence you can try a lot per connected client */
    if ((rc = SSL_connect(ssl)) <= 0)
    {
        scas_log (1, "%s: Error connecting SSL client object\n", __func__);
        if (print_ssl_error_msg (ssl, rc, 1, NULL))
        {
            SSL_free(ssl);
            return NULL;
        } 
        else
        {
            while (retry < 222)
            {
                if ((rc = SSL_connect(ssl)) <= 0)
                {
                    if (print_ssl_error_msg (ssl, rc, 0, NULL))
                    {
                        break; 
                    } 
                    else
                    {
                        /* According to the SSL error, you should try again. Introducing and I/O micro sleep */
                        retry++;
                        continue;
                    }
                }
                else
                {
                    /* Success */
                    scas_log_debug (5, "%s: SSL Connection opened\n", __func__);
                    return ssl;
                }
            }

            /* Non-succesfull / failure condition */
            /* If here, then the SSL connection isn't meant to be, so bailing out */
            print_ssl_error_msg (ssl, rc, 1, NULL);
            SSL_free(ssl);
            return NULL;
        }
    }
    else
    {
        /* Success */
        scas_log_debug (5, "%s: SSL Connection opened\n", __func__);
        return ssl;
    }
}


/******************************************************************************
Function:     SSL_connect_shutdown
Description:
              The SSL object will be shutdown fully

Parameters:
Returns:
        0 : Perfect shutdown
        1 : FUBAR
******************************************************************************/
int SSL_connect_shutdown (SSL * ssl)
{
    int retry = 0;
    int rc = 0;

    if (!ssl)
    {
        scas_log_debug (5, "Shutting down SSL; nothing to shutdown though... empty SSL pointer presented.\n");
        return 1;
    }
    else
    {
        scas_log_debug (5, "Shutting down SSL\n");
        while (retry < 10)
        {
          switch (rc = SSL_shutdown (ssl))
          {
            /* First SSL_shutdown must return 0 */
            /* Second SSL_shutdown must return 1 */
            case 0 :
                if (retry == 0)
                    scas_log_debug (5, "First shutdown succeeded, close notify send ok\n");
                else
                    scas_log_debug (2, "Retrying shutdown procedure (try: %d)\n", retry + 1);

                retry++;

                /* Perform microsleep to cut the system some slack to catch up */
                usleep (500 + 500 * retry); 
                continue;
            case 1 :
                scas_log_debug (2, "Succesfull shutdown #1\n");

                scas_log_debug (5, "Freeing SSL BIO's\n");
                BIO_free( SSL_get_rbio( ssl ) );

                SSL_CTX_sess_set_remove_cb(SSL_get_SSL_CTX( ssl ), NULL);
                /* SSL_CTX_free( SSL_get_SSL_CTX( ssl ) ); */

                scas_log_debug (2, "Succesfull shutdown #2\n");
                return 0;
            default :
                scas_log (3, "Shutdown failed completely, retry count is: %d\n", retry + 1);
                print_ssl_error_msg (ssl, rc, 1, NULL);

                scas_log_debug (5, "Freeing SSL BIO's\n");
                BIO_free( SSL_get_rbio( ssl ) );

                SSL_CTX_sess_set_remove_cb(SSL_get_SSL_CTX( ssl ), NULL);
                /* SSL_CTX_free( SSL_get_SSL_CTX( ssl ) ); */

                return 1;
           }
        }
    }

    /* Shouldn't get here */
    return 1;
}




/******************************************************************************
Function:   setup_client_ctx
Description:
    Setting up the SSL client side context with the right settings

Parameters:
Returns:
    >NULL : A pointer to ready to use SSL_CTX context pointer
    NULL  : FUBAR
******************************************************************************/
SSL_CTX *setup_client_ctx(char * cafile, 
                          char * cadir, 
                          char * clientcert, 
                          char * clientkey)
{
    SSL_CTX *ctx;

    scas_log_debug (5, "Entering: %s\n", __func__);
    /* ctx = SSL_CTX_new(SSLv23_method()); */
    ctx = SSL_CTX_new(SSLv23_client_method());

    if (!ctx)
    {
        scas_log (1, "%s: Error: No SSL context was created\n", __func__);
        return NULL;
    }


    if (SSL_CTX_load_verify_locations(ctx, cafile, cadir) != 1)
    {
        scas_log (1, "%s: Error loading CA file and/or directory", __func__);
        SSL_CTX_free (ctx);
        return NULL;
    }

    if (SSL_CTX_use_certificate_chain_file(ctx, clientcert) != 1) 
/*    if (SSL_CTX_use_certificate_file(ctx, clientcert, SSL_FILETYPE_PEM) != 1) */
/*         scas_log (1, "%s: Error loading certificate from file", __func__); */
    {
        scas_log (1, "%s: Error loading certificate chain from file", __func__);
        SSL_CTX_free (ctx);
        return NULL;
    }

    if (SSL_CTX_use_PrivateKey_file(ctx, clientkey, SSL_FILETYPE_PEM) != 1)
    {
        scas_log (1, "%s: Error loading private key from file", __func__);
        SSL_CTX_free (ctx);
        return NULL;
    }

    /*
     * Go into verify_callback after default SSL checks 
     * Perform signature checks (peer must present certificate and must verify it)
     */
#ifndef DISABLE_PROXY_SUPPORT
    SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER, scas_verify_callback);
    ctx->cert_store->check_issued = grid_check_issued_wrapper;

  #if OPENSSL_VERSION_NUMBER < 0x00908000L
    X509_STORE_set_flags(ctx->cert_store, X509_V_FLAG_CRL_CHECK |
                                          X509_V_FLAG_CRL_CHECK_ALL );
  #else
    X509_STORE_set_flags(ctx->cert_store, X509_V_FLAG_CRL_CHECK |
                                          X509_V_FLAG_CRL_CHECK_ALL |
                                          X509_V_FLAG_ALLOW_PROXY_CERTS );
  #endif /* OPENSSL_VERSION_NUMBER < 0x00908000L */
#endif /* DISABLE_PROXY_SUPPORT */


    /* Set max Chain depth */
    SSL_CTX_set_verify_depth(ctx, MAXCHAINDEPTH);
    SSL_CTX_set_options(ctx, SSL_OP_ALL|SSL_OP_NO_SSLv2);
    if (SSL_CTX_set_cipher_list(ctx, CIPHER_LIST) != 1)
    {
        scas_log (1, "%s: Error setting cipher list (no valid ciphers)", __func__);
        SSL_CTX_free (ctx);
        return NULL;
    }

    return ctx;
}


/******************************************************************************
Function:   setup_server_ctx
Description:
    Setting up the SSL server side context with the right settings

Parameters:
Returns:
    >NULL : A pointer to ready to use SSL_CTX context pointer    
    NULL  : FUBAR
******************************************************************************/
SSL_CTX *setup_server_ctx(char * cafile,
                          char * cadir,
                          char * servercert,
                          char * serverkey)
{
    SSL_CTX *ctx;

    scas_log_debug (5, "%s: creating context\n", __func__);
    ctx = SSL_CTX_new(SSLv23_server_method());
    if (!ctx)
    {
        scas_log (1, "%s: Error: No SSL context was created\n", __func__);
        return NULL;
    }
    else
    {
        scas_log_debug (5, "%s: Context created\n", __func__);
    }


    if (SSL_CTX_use_PrivateKey_file(ctx, serverkey, SSL_FILETYPE_PEM) != 1)
    {
        scas_log (1, "%s: Error loading private key from file '%s'\n", __func__, serverkey);
        SSL_CTX_free (ctx);
        return NULL;
    }
    else
    { 
        scas_log_debug (5, "%s: Private Key added to ctx\n", __func__);
    }


    /* 
     * Don't set the default OpenSSL verify paths with SSL_CTX_set_default_verify_paths(ctx).
     * Only the configured (Grid) trust anchors are trusted
     */
    if (SSL_CTX_load_verify_locations(ctx, cafile, cadir) != 1)
    {
        scas_log (1, "%s: Error loading CA file and/or directory '%s'\n", __func__, cafile);
        SSL_CTX_free (ctx);
        return NULL;
    }
    else
    { 
        scas_log_debug (5, "%s: CA directory added to ctx\n", __func__);
    }

    /* if (SSL_CTX_use_certificate_file(ctx, servercert, SSL_FILETYPE_PEM) != 1) */
    if (SSL_CTX_use_certificate_chain_file(ctx, servercert) != 1) 
    {
        scas_log (1, "%s: Error loading certificate from file '%s'\n", __func__, servercert);
        SSL_CTX_free (ctx);
        return NULL;
    }
    else
    { 
        scas_log_debug (5, "%s: Public certificate chain added to ctx\n", __func__);
    }


    /*
     * Go into verify_callback after default SSL checks 
     * Perform signature checks (peer must present certificate, must verify peer
     */


    /* SSL_CTX_set_purpose(ctx, X509_PURPOSE_ANY); */
    SSL_CTX_set_purpose(ctx, X509_PURPOSE_SSL_CLIENT);
/*  X509_STORE_CTX_set_purpose( verify_ctx, X509_PURPOSE_SSL_CLIENT ); */

    ctx->cert_store->check_issued = grid_check_issued_wrapper;
    SSL_CTX_set_verify(ctx, SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT, scas_verify_callback);
 
  #if OPENSSL_VERSION_NUMBER < 0x00908000L
    X509_STORE_set_flags(SSL_CTX_get_cert_store(ctx), X509_V_FLAG_CRL_CHECK |
                                                      X509_V_FLAG_CRL_CHECK_ALL );
  #else
    X509_STORE_set_flags(SSL_CTX_get_cert_store(ctx), X509_V_FLAG_CRL_CHECK |
                                                      X509_V_FLAG_CRL_CHECK_ALL |
                                                      X509_V_FLAG_ALLOW_PROXY_CERTS );
  #endif /* OPENSSL_VERSION_NUMBER < 0x00908000L */


#if 0 /* Should re-check this flag's placement. Normally found in or before the verify function set to the individual certificate from an SSL handshake. */
  #if OPENSSL_VERSION_NUMBER >= 0x00908000L
    cert->ex_flags |= EXFLAG_PROXY;
  #endif /* OPENSSL_VERSION_NUMBER >= 0x00908000L */
#endif

    /* Max depth */
    SSL_CTX_set_verify_depth(ctx, MAXCHAINDEPTH);



/**************************************************************************************/


    /* All (SSLv3, TLS) but not SSLv2 */
    SSL_CTX_set_options(ctx, SSL_OP_ALL | SSL_OP_NO_SSLv2);

    if (SSL_CTX_set_cipher_list(ctx, CIPHER_LIST) != 1)
    {
        scas_log (1, "%s: Error setting cipher list (no valid ciphers)\n", __func__);
        SSL_CTX_free (ctx);
        return NULL;
    }

    /* Does the Private key match the certificate? */
    if (SSL_CTX_check_private_key(ctx) != 1)
    {
        scas_log (1, "%s: Error matching private key\n", __func__);
        SSL_CTX_free (ctx);
        return NULL;
    }

    return ctx;
}




/******************************************************************************
Function:   x509IsCA
Description:
    Tests if the X509 * cert is a CA certificate or not
    => Example from GridSite
Parameters:
    A X509 pointer
Returns:
    0      : Not a CA cert
    1      : This is a CA cert
******************************************************************************/
int x509IsCA (X509 *cert)
{

/*
#define NEW_CA_SEARCH_METHOD
*/
#ifndef NEW_CA_SEARCH_METHOD
    int purpose_id;

    purpose_id = X509_PURPOSE_get_by_sname("sslclient");

    /* final argument to X509_check_purpose() is whether to check for CAness */

    if (X509_check_purpose(cert, purpose_id + X509_PURPOSE_MIN, 1))
        return 1;
    else return 0;
#else
    int                                 element_index = -1;
    int                                 critical;
    BASIC_CONSTRAINTS *                 x509v3_bc = NULL;

    if((x509v3_bc = X509_get_ext_d2i(cert,
                                     NID_basic_constraints,
                                     &critical,
                                     &element_index)) && x509v3_bc->ca)
    {
        /* Free allocated space */
        BASIC_CONSTRAINTS_free(x509v3_bc);

        /* CA cert found */ 
        return 1;
    }

    /* Not a CA cert */
    return 0;
#endif /* NEW_CA_SEARCH_METHOD */
}




/******************************************************************************
Function:       init_OpenSSL()
Description:    Do the SSL library initialization
Parameters:     void
Returns:
     0 : all is ok
     1 : failure reported must not continue with the SSL library
******************************************************************************/
int init_OpenSSL(void)
{
/*
    if (!THREAD_setup() || !SSL_library_init())
 */

    if (openssl_is_initialized)
        return 0;
        
    if (!SSL_library_init())
    {
        scas_log (1, "%s: ** OpenSSL initialization failed!\n", __func__);
        return 1;
    }
    SSL_load_error_strings();
    OpenSSL_add_ssl_algorithms();

    /* Declare OpenSSL to be initialized */
    openssl_is_initialized = 1;

    return 0;
}


/******************************************************************************
Function:       scas_verify_callback()
Description:    Execute the OpenSSL verify callback
                The ok will be the result of the default verification function
Parameters:     ok, X509_STORE_CTX *
Returns:        int
******************************************************************************/
int scas_verify_callback(int ok, X509_STORE_CTX *store_ctx)
{
    unsigned long   errnum   = X509_STORE_CTX_get_error(store_ctx);
    int             errdepth = X509_STORE_CTX_get_error_depth(store_ctx);
    X509           *cert     = X509_STORE_CTX_get_current_cert(store_ctx);
    STACK_OF(X509) *certstack = NULL;
    char *          cert_DN   = NULL;
    char *          issuer_DN = NULL;

    scas_log (5, "%s\n", __func__);

    cert_DN   = X509_NAME_oneline (X509_get_subject_name (cert), NULL, 0);
    issuer_DN = X509_NAME_oneline (X509_get_issuer_name (cert), NULL, 0);
    scas_log (5, "%s: DN:        %s\n", __func__, cert_DN);
    scas_log (5, "%s: Issuer DN: %s\n", __func__, issuer_DN);

    /* Resolve known error state and flag them as 'ok' */
    if (ok != 1)
    {
        if (errnum == X509_V_ERR_INVALID_CA) ok=1;
        if (errnum == X509_V_ERR_UNABLE_TO_GET_CRL) ok=1;
#if OPENSSL_VERSION_NUMBER >= 0x00908000L
        /* I don't want to do this, really, but I have yet to figure out
           how to get openssl 0.9.8 to accept proxy certificates when
           opening a new SSL connection...
        */
        if (errnum == X509_V_ERR_INVALID_PURPOSE) ok=1;
#endif
        if (errnum == X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION)
        {
            /* errnum = grid_knownCriticalExts( cert ); */
            errnum = grid_X509_knownCriticalExts (cert);
            store_ctx->error = errnum;
            if (errnum == X509_V_OK) ok=1;
        }
    }

    /*
     * We've now got the last certificate - the identity being used for
     * this connection. At this point we check the whole chain for valid
     * CAs or, failing that, GSI-proxy validity using grid_verifyProxy
     */
    if ( (errdepth == 0) && (ok == 1) )
    {
        certstack = (STACK_OF(X509) *) X509_STORE_CTX_get_chain (store_ctx);

        scas_log (5, "%s: End of certificate chain: performing full chain check in grid_verifyProxy()\n", __func__);

        errnum = grid_verifyProxy (certstack);
        store_ctx->error = errnum;
        if (errnum == X509_V_OK) ok=1;
    }

    if (ok != 1)
    {
        scas_log (2, "scas_verify_callback: error message = %s\n", 
                 X509_verify_cert_error_string (errnum));
    }


    /* The current way of working needs freeing */
    free (cert_DN);
    free (issuer_DN);

    return ok;
}


/******************************************************************************
Function:   grid_verifyCert
Description:
    Validates a certificate with the CRL list and the signing CA
******************************************************************************/
unsigned long grid_verifyCert( char * CA_DIR, STACK_OF(X509) *certstack )
{
    X509_STORE     *store      = NULL;
    X509_LOOKUP    *lookup     = NULL;
    X509_STORE_CTX *verify_ctx = NULL;
    X509           *cert       = NULL;
    char *          cert_DN    = NULL;
    char *          issuer_DN  = NULL;
    int             i = 0;
    int             depth = sk_X509_num( certstack );

    scas_log (5, "--- %s() start", __func__);

    /* Initials must be good */
    if ( CA_DIR == NULL )
    {
        scas_log (1, "Verifying certificate chain: No CA certificate directory specified." );
        return X509_V_ERR_APPLICATION_VERIFICATION;
    }
    if ( certstack == NULL )
    {
        scas_log (1, "Verifying certificate chain: Certificate stack is empty." );
        return X509_V_ERR_APPLICATION_VERIFICATION;
    }

    scas_log (2, "Using CA Directory: %s", CA_DIR);

    if (!(store = X509_STORE_new()))
    {
        scas_log (1, "Verifying certificate chain: Could not create a X509 STORE." );
        return ERR_get_error();
    }

    /* this is a macro, so we don't create a '' version */
    X509_STORE_set_verify_cb_func (store, scas_verify_callback);

    /* Executing the lookups to the CA and CRL files */
    if (X509_STORE_load_locations (store, NULL, CA_DIR) != 1)
    {
        scas_log (1, "Verifying certificate chain: Could not load the CA directory.");
        return ERR_get_error();
    }
    if (X509_STORE_set_default_paths(store) != 1)
    {
        scas_log (1, "Verifying certificate chain: Could not load the system wide CA certificates.");
        return ERR_get_error();
    }

    if (!(lookup = X509_STORE_add_lookup (store, X509_LOOKUP_hash_dir())))
    {
        scas_log (1, "Verifying certificate chain: Could not create X509_LOOKUP object.");
        return ERR_get_error();
    }

    if (!(X509_LOOKUP_add_dir (lookup, CA_DIR, X509_FILETYPE_PEM)))
    {
        scas_log (1, "Verifying certificate chain: Could not add CA_DIR.");
        return ERR_get_error();
    }

    store->check_issued = grid_check_issued_wrapper;
#if OPENSSL_VERSION_NUMBER < 0x00908000L
    X509_STORE_set_flags( store, X509_V_FLAG_CRL_CHECK |
                                 X509_V_FLAG_CRL_CHECK_ALL );
#else
    X509_STORE_set_flags( store, X509_V_FLAG_CRL_CHECK |
                                 X509_V_FLAG_CRL_CHECK_ALL |
                                 X509_V_FLAG_ALLOW_PROXY_CERTS );
#endif

    /* Creating a verification context and init it */
    if (!(verify_ctx = X509_STORE_CTX_new()))
    {
        scas_log (1, "Verifying certificate chain: Could not create a X509 STORE CTX (context).");
        return ERR_get_error();
    }

    for (i=depth-1; i >= 0; --i) 
    {
        if ((cert = sk_X509_value(certstack, i)))
        {
            cert_DN   = X509_NAME_oneline (X509_get_subject_name (cert), NULL, 0);
            issuer_DN = X509_NAME_oneline (X509_get_issuer_name (cert), NULL, 0);

            scas_log (5, "DN[%d]:        %s\n", i, cert_DN );
            scas_log (5, "Issuer DN[%d]: %s\n", i, issuer_DN);

            /* if (grid_IsCA(cert)) */
            if (x509IsCA(cert))
            {
                scas_log (5, "This certificate is a CA certificate\n" );
                scas_log (5, "     continue search for user certificate...\n" );
            }
            else
            {
                break;
            }

            free(cert_DN);
            free(issuer_DN);
        }
    }

    cert = sk_X509_value( certstack, 0 );

    cert_DN   = X509_NAME_oneline( X509_get_subject_name( cert ) , NULL, 0 );
    issuer_DN = X509_NAME_oneline( X509_get_issuer_name( cert )  , NULL, 0 );

    scas_log (5, "Certificate to verify:" );
    scas_log (5, "  DN:        %s", cert_DN );
    scas_log (5, "  Issuer DN: %s", issuer_DN );

    free(cert_DN);
    free(issuer_DN);


    if ( X509_STORE_CTX_init( verify_ctx, store, cert, certstack) != 1 )
    {
        scas_log (1, "Verifying certificate chain: Could not initialize verification context.");
        return ERR_get_error();
    }

    X509_STORE_CTX_set_purpose( verify_ctx, X509_PURPOSE_SSL_CLIENT );
#if OPENSSL_VERSION_NUMBER >= 0x00908000L
    cert->ex_flags |= EXFLAG_PROXY;
#endif

    if ( ( X509_verify_cert( verify_ctx ) ) != 1 )
    {
        return verify_ctx->error;
    }
    else
    {
        scas_log (2, "The verification of the certicate has succeeded.");
    }

    if ( verify_ctx ) X509_STORE_CTX_free( verify_ctx );
    if ( store )      X509_STORE_free( store );

    return X509_V_OK;
}




/******************************************************************************
Function:       grid_X509_knownCriticalExts
Description:    Check if the Critical Extention known to proxy certificates
                are set and understood.
Parameters:     X509 * cert
Returns:        unsigned long : X509_V_OK (good), X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION (bad)
******************************************************************************/
#define PROXYCERTINFO_OID      "1.3.6.1.5.5.7.1.14"
#define OLD_PROXYCERTINFO_OID  "1.3.6.1.4.1.3536.1.222"

unsigned long grid_X509_knownCriticalExts(X509 *cert)
{
   int  i = 0;
   char critical[80];
   X509_EXTENSION *ex;
   char            data[256];


   scas_log_debug (5, "Entering %s\n", __func__);
   for (i = 0; i < X509_get_ext_count(cert); ++i)
   {
        X509_NAME_oneline(X509_get_issuer_name(cert), data, 256);
        scas_log (5, "%s:  issuer  = %s\n", __func__, data);
        X509_NAME_oneline(X509_get_subject_name(cert), data, 256);
        scas_log (5, "%s:  subject = %s\n", __func__, data);
        ex = X509_get_ext(cert, i);

        if (X509_EXTENSION_get_critical(ex) &&
                                 !X509_supported_extension(ex))
        {
            OBJ_obj2txt(critical, sizeof(critical), X509_EXTENSION_get_object(ex), 1);

            scas_log (5, "%s: Critical extension found: %s", __func__, critical);
            if (strcmp(critical, PROXYCERTINFO_OID) == 0)     return X509_V_OK;
            if (strcmp(critical, OLD_PROXYCERTINFO_OID) == 0) return X509_V_OK;

            return X509_V_ERR_UNHANDLED_CRITICAL_EXTENSION;
        }
   }
   return X509_V_OK;
}


/******************************************************************************
Function:   grid_verifyProxy
Description:
    Tries to verify the proxies in the certstack
******************************************************************************/
unsigned long grid_verifyProxy( STACK_OF(X509) *certstack )
{
    int          i = 0;
    X509       * cert = NULL;
    time_t       now = time((time_t *)NULL);
    size_t       len = 0;             /* Lengths of issuer and cert DN */
    size_t       len2 = 0;            /* Lengths of issuer and cert DN */
    int          prevIsLimited = 0;   /* previous cert was proxy and limited */
    char         cert_DN   [256];     /* Pointer to current-certificate-in-certstack's DN */
    char         issuer_DN [256];     /* Pointer to issuer-of-current-cert-in-certstack's DN */
    char       * proxy_part_DN = NULL;
    int          depth = sk_X509_num (certstack);
    int          amount_of_CAs = 0;

    scas_log_debug (5, "%s: --- Welcome to the grid_verifyProxy function ---\n", __func__);

    if (depth == 0)
    {
        scas_log (1, "%s: Error: Empty certificate chain presented!\n", __func__);
        return X509_V_ERR_APPLICATION_VERIFICATION;
    }


    /* And there was (current) time... */
    time(&now);

    /* How many CA certs are there in the certstack? */
    for (i = 0; i < depth; i++)
    {
        if (x509IsCA(sk_X509_value(certstack, i)))
            amount_of_CAs++;
    }


    scas_log_debug (5, "%s: #CA's = %d , depth = %d\n", __func__, amount_of_CAs, depth);

    /* Check if our certificate chain could hold anything useful, like proxies or EECs */
    if ((amount_of_CAs + 2) > depth)
    {
        if ((depth - amount_of_CAs) > 0)
        {
            scas_log (1, "%s: No proxy certificate in certificate stack to check.\n", __func__);
            return X509_V_OK;
        }
        else
        {
            scas_log (1, "%s: No personal certificate (neither proxy or user certificate) found in the certficiate stack.", __func__);
            return X509_V_ERR_APPLICATION_VERIFICATION;
        }
    }


    /* 
     * Changed this value to start checking the proxy and such and
     * to skip the CA and the user_cert
     */

    for (i = depth - (amount_of_CAs + 2); i >= 0; i--)
    {
        /* Check for X509 certificate and point to it with 'cert' */
        if ( (cert = sk_X509_value(certstack, i)) != NULL )
        {
            X509_NAME_oneline(X509_get_subject_name(cert), cert_DN,   256);
            X509_NAME_oneline(X509_get_issuer_name(cert),  issuer_DN, 256);
            len       = strlen( cert_DN );
            len2      = strlen( issuer_DN );

            scas_log_debug (5, "%s: Proxy to verify:\n", __func__ );
            scas_log_debug (5, "%s:   Issuer DN: %s\n", __func__, issuer_DN );
            scas_log_debug (5, "%s:   DN:        %s\n", __func__, cert_DN );

            if (now < grid_asn1TimeToTimeT((char *)ASN1_STRING_data(X509_get_notBefore(cert)),0))
            {
                scas_log (1, "%s: Proxy certificate is not yet valid.\n", __func__);
                return X509_V_ERR_CERT_NOT_YET_VALID;
            }

            if (now > grid_asn1TimeToTimeT((char *)ASN1_STRING_data(X509_get_notAfter(cert)),0))
            {
                scas_log (1, "%s: Proxy certificate expired.\n", __func__);
                return X509_V_ERR_CERT_HAS_EXPIRED;
            }

            /* User not allowed to sign shortened DN */
            /* Note: Only true for GT2 proxies */
            if (len2 > len)
            {
                scas_log (1, "%s: It is not allowed to sign a shorthened DN.\n", __func__);
                return X509_V_ERR_INVALID_CA;
            }

            /* Proxy subject must begin with issuer. */ /* This is not true in RFC/GT4 proxies */
            if (strncmp(cert_DN, issuer_DN, len2) != 0)
            {
                scas_log (1, "%s: Proxy subject must begin with the issuer.\n", __func__);
                return X509_V_ERR_INVALID_CA;
            }

            /* Set pointer to end of base DN in cert_DN */
            proxy_part_DN = &cert_DN[len2];

            /* First attempt at support for Old and New style GSI
               proxies: /CN=anything is ok for now */
            if (strncmp(proxy_part_DN, "/CN=", 4) != 0)
            {
                scas_log (1, "%s: Could not find a /CN= structure in the DN, thus it is not a proxy.\n", __func__);
                return X509_V_ERR_INVALID_CA;
            }

            /* The following structures need to be reworked to support for GSI/RFC3820 proxies */
            if ((strncmp(proxy_part_DN, "/CN=proxy", 9) == 0) && (i > 0))
            {
                scas_log (3, "%s:  Found old style proxy.\n", __func__);
            }

            if ((strncmp(proxy_part_DN, "/CN=limited proxy", 17) == 0) && (i > 0))
            {
                prevIsLimited = 1;
                scas_log (3, "%s:  Found old style limited proxy.\n", __func__);
            }
            else
            {
                if (prevIsLimited)
                {
                    scas_log (1, "%s: Proxy chain integrity error. Previous proxy in chain was limited, but this one is a regular proxy.\n", __func__);
                    return X509_V_ERR_INVALID_CA;
                }
            }
            scas_log_debug (5, "%s:   Proxy is valid\n", __func__);
        }
    }

    return X509_V_OK;
}


/**
 * Note that timegm() is non-standard. Linux manpage advices the following
 * substition instead.
 */
time_t my_timegm(struct tm *tm)
{
   time_t ret;
   char *tz;

   tz = getenv("TZ");
   setenv("TZ", "", 1);
   tzset();
   ret = mktime(tm);
   if (tz)
       setenv("TZ", tz, 1);
   else
       unsetenv("TZ");
   tzset();

   return ret;
}



/**
 *  (Use ASN1_STRING_data() to convert ASN1_GENERALIZEDTIME to char * if
 *   necessary)
 */

time_t grid_asn1TimeToTimeT(char *asn1time, size_t len)
{
   char   zone;
   struct tm time_tm;

   if (len == 0) len = strlen(asn1time);

   if ((len != 13) && (len != 15)) return 0; /* dont understand */

   if ((len == 13) &&
       ((sscanf(asn1time, "%02d%02d%02d%02d%02d%02d%c",
         &(time_tm.tm_year),
         &(time_tm.tm_mon),
         &(time_tm.tm_mday),
         &(time_tm.tm_hour),
         &(time_tm.tm_min),
         &(time_tm.tm_sec),
         &zone) != 7) || (zone != 'Z'))) return 0; /* dont understand */

   if ((len == 15) &&
       ((sscanf(asn1time, "20%02d%02d%02d%02d%02d%02d%c",
         &(time_tm.tm_year),
         &(time_tm.tm_mon),
         &(time_tm.tm_mday),
         &(time_tm.tm_hour),
         &(time_tm.tm_min),
         &(time_tm.tm_sec),
         &zone) != 7) || (zone != 'Z'))) return 0; /* dont understand */

   /* time format fixups */

   if (time_tm.tm_year < 90) time_tm.tm_year += 100;
   --(time_tm.tm_mon);

   return my_timegm(&time_tm);
}




int grid_check_issued_wrapper(X509_STORE_CTX *ctx,X509 *x,X509 *issuer)
/* We change the default callback to use our wrapper and discard errors
   due to GSI proxy chains (ie where users certs act as CAs) */
{
    int ret = 0;
    char * cert_DN   = NULL;
    char * issuer_DN = NULL;

    cert_DN   = X509_NAME_oneline(X509_get_subject_name(x), NULL, 0);
    issuer_DN = X509_NAME_oneline(X509_get_issuer_name(issuer), NULL, 0);

    scas_log_debug (3, "%s:  issuer   = %s\n", __func__, issuer_DN);
    scas_log_debug (3, "%s:  subject  = %s\n", __func__, cert_DN);

    free(cert_DN);
    free(issuer_DN);

    /* If all is ok, move out of here */
    if ((ret = X509_check_issued(issuer, x)) == X509_V_OK)
        return 1;
 
    scas_log (5, "%s: check_issued returned: %s\n", __func__, X509_verify_cert_error_string (ret));
 
    /* Non self-signed certs without signing are ok if they passed
           the other checks inside X509_check_issued. Is this enough? */
    if ((ret == X509_V_ERR_KEYUSAGE_NO_CERTSIGN) &&
        (X509_subject_name_cmp(issuer, x) != 0)) return 1;
 
    /* If we haven't asked for issuer errors don't set ctx */
#if OPENSSL_VERSION_NUMBER < 0x00908000L
    if (!(ctx->flags & X509_V_FLAG_CB_ISSUER_CHECK)) return 0;
#else
    if (!(ctx->param->flags & X509_V_FLAG_CB_ISSUER_CHECK)) return 0;
#endif
 
    ctx->error = ret;
    ctx->current_cert = x;
    ctx->current_issuer = issuer;

    scas_log (5, "%s: calling ctx->verify_cb\n", __func__);
    return ctx->verify_cb(0, ctx);
}
    


/******************************************************************************
Function:       post_connection_check
Description:    Post Connection Checking (only for client)
                DNS (or final /CN=<host>) 
                If SubjectAltNames exist, then they must match the host. If they don't exist, then look at the final CN field in the subject.
                All else is a failure
Parameters:     SSL *, char * host
Returns:        long
******************************************************************************/
long post_connection_check(SSL *ssl, const char *host)
{
    X509      *      peer_cert = NULL;
    X509_NAME *      subj;
    char             buf_data[256];
    int              extcount;
    int              ok = 0;
    int              found_DNS_entry = 0;

    scas_log_debug (5, "%s\n", __func__);

    if (!host)
    {
        scas_log (1, "%s: error: No hostname supplied to check\n", __func__);
        goto err_occured;
    }

    if (!(peer_cert = SSL_get_peer_certificate(ssl)) || !host)
    {
        scas_log (1, "%s: error: Could not get peer certificate from the SSL handle for host: %s\n",
                     __func__,
                     host);
        goto err_occured;
    }
  

  #if OPENSSL_VERSION_NUMBER >= 0x00908000L
    peer_cert->ex_flags |= EXFLAG_PROXY;
  #endif /* OPENSSL_VERSION_NUMBER >= 0x00908000L */ 

    /* Compare the subjectAltName DNS value with the host value */
    if ((extcount = X509_get_ext_count(peer_cert)) > 0)
    {
        int i;

        for (i = 0; i < extcount; i++)
        {
            X509_EXTENSION    *ext;
            int NID_from_ext = NID_undef; /* Initialize with undefined NID (Numerical ID of a type of ASN1 object) */

            ext = X509_get_ext(peer_cert, i);
            NID_from_ext = OBJ_obj2nid(X509_EXTENSION_get_object(ext));

            if (NID_from_ext == NID_subject_alt_name)
            {
                int                  j;
                unsigned char        *data;
                STACK_OF(CONF_VALUE) *val;
                CONF_VALUE           *nval;
                X509V3_EXT_METHOD    *meth;
                void                 *ext_str = NULL;

                if (!(meth = X509V3_EXT_get(ext)))
                    break;
                data = ext->value->data;

#if (OPENSSL_VERSION_NUMBER > 0x00907000L)
                if (meth->it)
                    ext_str = ASN1_item_d2i(NULL, &data, ext->value->length,
                                            ASN1_ITEM_ptr(meth->it));
                else
                    ext_str = meth->d2i(NULL, &data, ext->value->length);
#else
                ext_str = meth->d2i(NULL, &data, ext->value->length);
#endif
                val = meth->i2v(meth, ext_str, NULL);
                for (j = 0;  j < sk_CONF_VALUE_num(val);  j++)
                {
                    nval = sk_CONF_VALUE_value(val, j);
                    if (!strcmp(nval->name, "DNS") && !strcasecmp(nval->value, host))
                    {
                        scas_log (5, "%s: subjectAltName DNS entry matches host\n", __func__);
                        ok = 1;
                        break;
                    }
                    scas_log_debug (5, "%s: found name: %s value: %s\n", __func__, nval->name, nval->value);
                    if (!strcmp(nval->name, "DNS"))
                    {
                        found_DNS_entry = 1;
                    }
                }

                /* Clean up ASN1 item */
                /* ASN1_item_free((ASN1_VALUE *)ext_str, 
                                  ASN1_ITEM_ptr(meth->it)); */
                /* Conf value removal */
                sk_CONF_VALUE_pop_free( val, X509V3_conf_free );
            }
            if (ok)
                break;
        }
    }


    if (found_DNS_entry && !ok)
    {
        scas_log (1, "%s: Error: found one or more DNS entries in the subjectAltName but none match the host '%s'.\n", __func__, host);
        goto err_occured;
    }


    /* If the subjectAltName DNS value doesn't match the host value, 
     * resort back to the X509v1 solution which is to compare the host value 
     * with the final CN value in the cert's subject
     */
    if (!ok && (subj = X509_get_subject_name(peer_cert)) &&
        X509_NAME_get_text_by_NID(subj, NID_commonName, buf_data, 256) > 0)
    {
        scas_log (5, "%s: No subjectAltName:DNS entry found, trying final CN field in subject to match contacted host.\n", __func__);
        if (strcasecmp(buf_data, host) != 0)
        {
            scas_log (1, "%s: Error: final CN field in subject '%s' and host '%s' do not match!\n", __func__, buf_data, host);
            goto err_occured;
        }
        else
        {
            scas_log (5, "%s: final CN field in subject matches host\n", __func__);
            ok = 1;
        }
    }


    if (peer_cert)
        X509_free(peer_cert);


    if (peer_cert)
        X509_free(peer_cert);

    if (ok)
        return X509_V_OK;
    else
        return X509_V_ERR_APPLICATION_VERIFICATION;

err_occured:
    if (peer_cert)
        X509_free(peer_cert);
    return X509_V_ERR_APPLICATION_VERIFICATION;
}



/******************************************************************************
Function:       x509_chain_to_dn()
Description:    Get the base DN of the X509 proxy cert
Parameters:
                certstack: certificate chain
Returns:        dn (which should be freed)
******************************************************************************/
char * x509_chain_to_dn (STACK_OF(X509) *certstack)
{
    int i = 0;
    int amount_of_CAs = 0;
    int depth = sk_X509_num(certstack);

    /* Do we have a chain at all... */
    if (depth > 0)
    {
        /* Amount of CAs in chain */
        for (i = 0; i < depth; i++)
        {
            if (x509IsCA(sk_X509_value(certstack, i)))
                amount_of_CAs++;
        }

        /* Select the User certificate */
        i = depth - (amount_of_CAs + 1);

        if (i < 0)
        {
            i = 0;
            return NULL;
        }

        /* This returned value must be free'd */
        return X509_NAME_oneline(X509_get_subject_name(sk_X509_value(certstack, i)),NULL,0);
    }
    else
        return NULL;
}



/******************************************************************************
Function:       seed_prng
Description:    seed prng with /dev/urandom
Parameters:     void
Returns:        void
******************************************************************************/
void seed_prng(void)
{
     RAND_load_file("/dev/urandom", 1024);
}




/****************************************************
Function: print_ssl_error_msg

Variables:
    SSL * ssl   :   pointer to SSL structure (for this connection)
    int ret     :   return code of the previously SSL command
    printmsg    :   0 == don't print the messages, 1 =< print the message

Return:
    0 : Non lethal failure, should consider a retry of the SSL interaction
    1 : Lethal failure, recoverable situation
****************************************************/
int print_ssl_error_msg (SSL * ssl, int ret, int printmsg, int * error_code )
{
    long rc = 0;
    int unrecoverable = 0;

    if (ssl)
    {
        rc = SSL_get_error (ssl, ret);
    }
    else
    {
        rc = ERR_get_error();
    }


    switch (rc)
    {
        case SSL_ERROR_NONE :
        {
            /* No error */
            if (printmsg)
                scas_log (5, "XACML SSL I/O: Got error trigger, but there's no error on the stack\n");
            unrecoverable = 0;
            break;
        }
        case SSL_ERROR_ZERO_RETURN :
        {
            /* Shutdown given, possibly not shutdown yet in the underlying transport layer */
            if (printmsg)
                scas_log (5, "XACML SSL I/O: Shutdown given, possibly not shutdown yet in the underlying transport layer\n");
            unrecoverable = 0;
            break;
        }
        case SSL_ERROR_WANT_READ:
        case SSL_ERROR_WANT_WRITE:
        {
            if (printmsg)
                scas_log (5, "XACML SSL I/O: operation not completed, retry that last I/O operation\n");
            unrecoverable = 0;
            break;
        }
        case SSL_ERROR_WANT_CONNECT:
        case SSL_ERROR_WANT_ACCEPT:
        {
            if (printmsg)
                scas_log (5, "XACML SSL I/O: operation not completed, underlying I/O not ready, retry that last accept/connect operation. Re-check select() or poll() logics\n");
            unrecoverable = 0;
            break;
        }
        case SSL_ERROR_WANT_X509_LOOKUP:
        {
            if (printmsg)
                scas_log (5, "XACML SSL I/O: operation not completed, application callback was called again, should retry.\n");
            unrecoverable = 0;
            break;
        }
        case SSL_ERROR_SYSCALL:
        {
            /* Some I/O error occurred. The OpenSSL error queue may contain more information on the error. If the error queue is empty (i.e. ERR_get_error() returns 0), ret can be used to find out more about the error: If ret == 0, an EOF was observed that violates the protocol. If ret == -1, the underlying BIO reported an I/O error (for socket I/O on Unix systems, consult errno for details). */
            if (printmsg)
                 scas_log (0, "XACML SSL I/O: Some I/O error occurred. Unrecoverable\n");
            unrecoverable = 1;
            break;
        }
        case SSL_ERROR_SSL:
        {
            /* Check ERR_get_error() for more information */
            if (printmsg)
                scas_log (0, "XACML SSL I/O: A failure in the SSL library occurred, usually a protocol error.\n");
            unrecoverable = 0;
            break;
        }

        default:
        {
            if (printmsg)
                scas_log (0, "XACML SSL I/O: Undefined error code, not in OpenSSL specs\n");
            unrecoverable = 1;
            break;
        }
    }

    if (error_code)
        *error_code = rc;


    return unrecoverable;
}

